# Copyright 2004 Tom Rothamel <tom-renpy@onegeek.org>
# 
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import re

import language
import fragment

class ScriptParser(object):


    # The lexer.

    def __init__(self, text):
        self.text = text
        self.pos = 0
        self.line = 1
        self.indents = [ -1 ]

    def state(self):
        """
        Returns an opaque object representing the current state of the
        lexer, that can be passed to backtrack().
        """

        return self.pos, self.line, self.indents[:]
    
    def backtrack(self, state):
        """
        When given an object returned from state(), backtracks the lexer
        to that state.
        """

        self.pos, self.line, self.indents = state

    def match(self, r):
        """
        If re matches at the current lexical position, returns the match
        object corresponding to re, and updates the current lexer position
        to be after the match. Otherwise, returns None and leaves the lexer
        position unchanged.

        Also updates self.line to reflect the new location in the file.
        """
        
        r = re.compile(r)
        m = r.match(self.text, self.pos)

        if m:
            self.pos = m.end()
            self.line += m.group(0).count("\n")

        return m


    def eat(self, r):
        """
        First repeatly matches all whitespace until no more can be matched,
        then tries to match the regular expression. It will always consume
        the whitespace, and will return a match object if the regular
        expression was matched.
        """

        while self.whitespace():
            pass

        return self.match(r)

    def error(self, msg):
        """
        Signals an error at the current position in the file.
        """

        raise Exception, "%s on line %d, before %r." % (msg,
                                                        self.line,
                                                        self.text[self.pos:self.pos + 16]
                                                        )

    def need(self, what):
        """
        What may be a function or a regular expression. Tries to match
        it at the current location, and throws an error if it cant.
        """

        if callable(what):
            m = what()
        else:
            m = self.eat(what)
        
        if m is None:
            self.error("Syntax error")
        else:
            return m


    def debug(self):
        print repr(self.text[self.pos:self.pos + 50])

    # Magic.

    def magic_indent(self, mindent, indent, allow_more):
        """
        Returns true if the indentation after the current position is
        consistent with the indentation of the current block, and
        false otherwise.

        The following rules apply:

        - The end of the file is inconsistent with any level of
        indentation. 

        - A line consisting entirely of whitespace is always
        consistent, and the function will always return the current
        value of indent.

        Otherwise, the amount of indentation at the start of the line is
        computed, and the following rules are applied:

        - If indent is True, then a line is consistent if the
        indentation is greater than or equal to mindent spaces. The
        actual indentation is returned.

        - If indent is not True, then the line is consistent if the
        indentation of the line is equal to indent. The current value
        of indent is returned.

        If a line is inconsistent, False is returned.
        """

        m = re.compile(r'([ \t]*)(.|\n)').match(self.text, self.pos)

        if not m:
            return False

        # Whitespace ends with a newline or comment.
        if m.group(2) in ('#', '\n'):
            return indent

        n = len(m.group(1).expandtabs())
        
        if indent is True:
            if n <= mindent:
                return False
            else:
                return n
        else:
            if n < indent:
                return False
            elif n > indent and not allow_more:
                self.error("Indentation mismatch")
            else:
                return indent        

    def begin_indent(self):
        self.indents.append(True)

    def end_indent(self):
        self.indents.pop()

    def indent(self, allow_more=False):

        indent = self.magic_indent(self.indents[-2], self.indents[-1],
                                   allow_more)

        if indent is False:
            return False
        else:
            self.indents[-1] = indent
            return True

    def python_line(self, term="\n"):
        """
        This tries to parse a line of python code, starting at the
        current position. It automatically continues the line when
        it encounters newlines inside quotes or parenthesis. It
        consumes the newline terminating the line of python code.

        It doesn't support triple-quoted strings. It also doesn't
        check the python for synatactic correctness.
        """

        start = self.pos
        parens = 0

        while True:

            c = self.text[self.pos]
            self.pos += 1

            if c == "\n":
                self.line += 1

            if parens == 0 and c == term:
                break

            if c == "\\":
                self.pos += 1
                continue

            if c == '"' or c == "'":
                delim = c

                while True:
                    c = self.text[self.pos]
                    self.pos += 1

                    if c == delim:
                        break

                    if c == "\n":
                        self.line += 1

                    if c == "\\":
                        self.pos += 1

                continue

            if c in ('(', '[', '{'):
                parens += 1
                continue

            if c in (')', ']', '}'):
                parens -= 1
                continue

        return self.text[start:self.pos]
        
    # Tokens.

    def whitespace(self):
        return ( self.match(r'[ \t]+') or
                 self.match(r'\\\n') or
                 self.match(r'#.*') )

    def NL(self):
        return self.eat(r'\n')

    keywords = [
        "call",
        "character",
        "cmenu",
        "else",
        "hide",
        "if",
        "image",
        "include",
        "jump",
        "label",
        "menu",
        "pass",
        "python",
        "return",
        "seenset",
        "scene",
        "show",
        "with",
        ]

    def name(self):
        state = self.state()
        m = self.eat(r'\w+')

        if not m:
            return None

        n = m.group(0)
        if n in self.keywords:
            self.backtrack(state)
            return None

        return n

    def string(self):

        m = ( self.eat(r'"(([^"\\]|\\.)*)"') or
              self.eat(r"'(([^'\\]|\\.)*)'") )

        if not m:
            return None

        s = m.group(1)

        # Get rid of whitespace.
        s = re.sub(r"\s+", " ", s)

        replacements = {
            'n' : '\n',
            '\\' : '\\',
            't' : '    ',
            }


        def resub(m):
            c = m.group(1)
            if c in replacements:
                return replacements[c]
            else:
                return c

        s = s.strip()
        s = re.sub(r"\\(.)", resub, s)

        return s


    ######################################################################
    # Rules: Start and Common

    def start(self):
        decls = self.decl_block()
        block = self.block()

        # Check to be sure we're at the end of the string.

        if self.pos != len(self.text):
            self.error("Unexpected end of statements")

        return decls, block


    # The name of a displayable image/image like thing.
    def thing(self):

        ## ( name )+
        
        n = self.name()

        if not n:
            return None

        rv = [ n ]

        while True:
            n = self.name()
            if not n:
                break

            rv.append(n)

        return tuple(rv)
        

    # Python lines and blocks.
    def python(self):

        if self.eat(r"\$"):
            code = self.python_line().strip()
            return fragment.Python(code)

        if self.eat(r"python"):
            self.need(":")
            self.need(self.NL)

            self.begin_indent()

            code = "try:\n"

            while self.indent(True):
                code += self.python_line()

            self.end_indent()

            code += "finally:\n"
            code += "    pass\n"

            return fragment.Python(code)

        return None

    ######################################################################
    # Rules: Declarations.

    def decl_block(self):

        ## at the same indent level, any number of:

        self.begin_indent()

        rv = [ fragment.Pass() ]

        while True:

            state = self.state()

            ##    <empty line>
            if self.NL():
                continue

            ##    "include" string
            if self.eat("include"):
                target = self.need(self.string)
                self.need(self.NL)

                language.parse_file(target)
                continue

            ##    declaration
            d = self.declaration()

            if d:
                rv.append(d)
                continue

            break

        self.end_indent()
        return rv

    def declaration(self):

        ## "character" name
        if self.eat("character"):
            name = self.need(self.name)
            self.need(self.NL)

            return fragment.CharacterDeclaration(name)

        ## "image" thing ":" python_line 
        if self.eat("image"):
            thing = self.need(self.thing)
            self.need(":")
            code = self.python_line().strip()
            
            return fragment.ImageDeclaration(thing, code)

        ## python
        python = self.python()
        if python:
            return python

        return None

    ######################################################################
    # Rules: Declarations.

    def block(self, nonempty=True):

        rv = [ ]

        self.begin_indent()

        ## at the same indentation level:
        while self.indent():

            ##     <empty line>
            if self.NL():
                continue

            ##     "label" name ":" NL block
            if self.eat("label"):
                name = self.need(self.name)
                self.need(":")
                self.need(self.NL)

                rv.append(fragment.Label(name))

                block = self.block(nonempty=False)
                rv.extend(block)

                continue

            ##    statement
            s = self.statement()
            rv.append(s)

        self.end_indent()

        if nonempty and not rv:
            self.error("Empty block not allowed")

        return rv

    def block_or_statement(self):

        # Newline followed by a block, or...
        if self.NL():
            return self.block()

        # ... a statement on the same line.
        else:
            return [ self.statement() ]
        

    def statement(self):

        # Python stuff.
        p = self.python()
        if p:
            return p

        ### control flow:

        ## if cond: sob [ else: sob ]
        if self.eat("if"):
            cond = self.python_line(':')[:-1].strip()
            true = self.need(self.block_or_statement)

            state = self.state()                

            if self.indent() and self.eat("else"):
                self.need(":")
                false = self.need(self.block_or_statement)
            else:
                self.backtrack(state)
                false = None

            return fragment.If(cond, true, false)

        ## jump name NL
        if self.eat("jump"):
            name = self.need(self.name)
            self.need(self.NL)

            return fragment.Jump(name)

        ## pass NL
        if self.eat("pass"):
            self.need(self.NL)

            return fragment.Pass()

        ## call name NL
        if self.eat("call"):
            name = self.need(self.name)
            self.need(self.NL)

            return fragment.Call(name)

        ## return NL
        if self.eat("return"):
            self.need(self.NL)

            return fragment.Return()

        ### Scene mangement.

        ## scene ( thing )? NL
        if self.eat("scene"):
            thing = self.thing()
            self.need(self.NL)

            return fragment.Scene(thing)

        ## show thing NL
        if self.eat("show"):
            thing = self.need(self.thing)
            self.need(self.NL)

            return fragment.Show(thing)

        ## hide thing NL
        if self.eat("hide"):
            thing = self.need(self.thing)
            self.need(self.NL)

            return fragment.Hide(thing)

        ## menu <menu body>
        if self.eat("menu"):
            return self.menu(False)

        ## cmenu <menu body>
        if self.eat("cmenu"):
            return self.menu(True)

        ## name string
        who = self.name()
        if who:
            what = self.need(self.string)
            self.need(self.NL)

            return fragment.Dialogue(who, what, True)

        ## string string
        ## string
        s = self.string()
        if s:
            s1 = self.string()
            self.need(self.NL)

            if s1:
                return fragment.Dialogue(s, s1, False)
            else:
                return fragment.Dialogue(None, s, False)

        self.error("Syntax error (not a statement)")
            
    def menu(self, centered):

        name = self.name()
        label = None
        choices = [ ]
        seenset = None

        self.need(":")
        self.need(self.NL)

        self.begin_indent()

        while self.indent():

            if self.NL():
                continue

            if self.eat("seenset"):
                if seenset:
                    self.error("A menu may have only one seenset")

                seenset = self.need(self.name)
                self.need(self.NL)

                continue
                
            s = self.need(self.string)

            if self.NL():
                if label:
                    self.error("A menu may have only one label")
                elif choices:
                    self.error("Menu label must occur before choices")
                else:
                    label = s
                    continue
                
            if self.eat("if"):
                cond = self.python_line(':')[:-1].strip()
            else:
                self.need(":")
                cond = None
                
            choices.append((s, cond, self.block_or_statement()))

        self.end_indent()
        
        return fragment.Menu(name, label, choices, centered, seenset)
